/***************************************************
  This is our library for the Adafruit ILI9341 Breakout and Shield
  ----> http://www.adafruit.com/products/1651

  Check out the links above for our tutorials and wiring diagrams
  These displays use SPI to communicate, 4 or 5 pins are required to
  interface (RST is optional)
  Adafruit invests time and resources providing this open source code,
  please support Adafruit and open-source hardware by purchasing
  products from Adafruit!

  Written by Limor Fried/Ladyada for Adafruit Industries.
  MIT license, all text above must be included in any redistribution
 ****************************************************/

#include "ILI9341_SPI.h"
#ifdef __AVR
  #include <avr/pgmspace.h>
#elif defined(ESP8266)
  #include <pgmspace.h>
#endif

#ifndef ARDUINO_STM32_FEATHER
  #include "pins_arduino.h"
  #include "wiring_private.h"
#endif

//#include <limits.h>
#include <SPI.h>

#define TFTWIDTH   240    
#define TFTHEIGHT  320

#define TFTREGISTER_SOFTRESET          0x01
#define TFTREGISTER_SLEEPIN            0x10
#define TFTREGISTER_SLEEPOUT           0x11
#define TFTREGISTER_NORMALDISP         0x13
#define TFTREGISTER_INVERTOFF          0x20
#define TFTREGISTER_INVERTON           0x21
#define TFTREGISTER_GAMMASET           0x26
#define TFTREGISTER_DISPLAYOFF         0x28
#define TFTREGISTER_DISPLAYON          0x29
#define TFTREGISTER_COLADDRSET         0x2A
#define TFTREGISTER_PAGEADDRSET        0x2B
#define TFTREGISTER_MEMORYWRITE        0x2C
#define TFTREGISTER_PIXELFORMAT        0x3A
#define TFTREGISTER_FRAMECONTROL       0xB1
#define TFTREGISTER_DISPLAYFUNC        0xB6
#define TFTREGISTER_ENTRYMODE          0xB7
#define TFTREGISTER_POWERCONTROL1      0xC0
#define TFTREGISTER_POWERCONTROL2      0xC1
#define TFTREGISTER_VCOMCONTROL1      0xC5
#define TFTREGISTER_VCOMCONTROL2      0xC7
#define TFTREGISTER_MEMCONTROL      0x36

#define TFTREGISTER_MEMCONTROL_MY  0x80
#define TFTREGISTER_MEMCONTROL_MX  0x40
#define TFTREGISTER_MEMCONTROL_MV  0x20
#define TFTREGISTER_MEMCONTROL_ML  0x10
#define TFTREGISTER_MEMCONTROL_RGB 0x00
#define TFTREGISTER_MEMCONTROL_BGR 0x08
#define TFTREGISTER_MEMCONTROL_MH  0x04


// If the SPI library has transaction support, these functions
// establish settings and protect from interference from other
// libraries.  Otherwise, they simply do nothing.
#ifdef SPI_HAS_TRANSACTION
static inline void spi_begin(void) __attribute__((always_inline));
static inline void spi_begin(void) {
#if defined (ARDUINO_ARCH_ARC32)
  // max speed!
  SPI.beginTransaction(SPISettings(16000000, MSBFIRST, SPI_MODE0));
#else
    // max speed!
  SPI.beginTransaction(SPISettings(24000000, MSBFIRST, SPI_MODE0));
#endif
}
static inline void spi_end(void) __attribute__((always_inline));
static inline void spi_end(void) {
  SPI.endTransaction();
}
#else
#define spi_begin()
#define spi_end()
#endif


// Constructor when using software SPI.  All output pins are configurable.
ILI9341_SPI::ILI9341_SPI(int8_t cs, int8_t dc, int8_t mosi,
				   int8_t sclk, int8_t rst, int8_t miso) : Adafruit_GFX(TFTWIDTH, TFTHEIGHT) {
  _cs   = cs;
  _dc   = dc;
  _mosi  = mosi;
  _miso = miso;
  _sclk = sclk;
  _rst  = rst;
  hwSPI = false;
  init();
}


// Constructor when using hardware SPI.  Faster, but must use SPI pins
// specific to each board type (e.g. 11,13 for Uno, 51,52 for Mega, etc.)
ILI9341_SPI::ILI9341_SPI(int8_t cs, int8_t dc, int8_t rst) : Adafruit_GFX(TFTWIDTH, TFTHEIGHT) {
  _cs   = cs;
  _dc   = dc;
  _rst  = rst;
  hwSPI = true;
  _mosi  = _sclk = 0;
  init();
}


void ILI9341_SPI::init(void) {
  if (_rst > 0) {
    pinMode(_rst, OUTPUT);
    digitalWrite(_rst, HIGH);
  }
  pinMode(_dc, OUTPUT);
  pinMode(_cs, OUTPUT);
  digitalWrite(_dc, LOW);
  digitalWrite(_cs, HIGH);


#if defined (USE_FAST_PINIO)
  csport    = portOutputRegister(digitalPinToPort(_cs));
  cspinmask = digitalPinToBitMask(_cs);
  dcport    = portOutputRegister(digitalPinToPort(_dc));
  dcpinmask = digitalPinToBitMask(_dc);
#endif

  if(hwSPI) { // Using hardware SPI
    SPI.begin();

#ifndef SPI_HAS_TRANSACTION
    SPI.setBitOrder(MSBFIRST);
    SPI.setDataMode(SPI_MODE0);
  #if defined (_AVR__)
    SPI.setClockDivider(SPI_CLOCK_DIV2); // 8 MHz (full! speed!)
    mySPCR = SPCR;
  #elif defined(TEENSYDUINO)
    SPI.setClockDivider(SPI_CLOCK_DIV2); // 8 MHz (full! speed!)
  #elif defined (__arm__)
    SPI.setClockDivider(11); // 8-ish MHz (full! speed!)
  #endif
#endif
  } 
  else {
    pinMode(_sclk, OUTPUT);
    pinMode(_mosi, OUTPUT);
    pinMode(_miso, INPUT);
    digitalWrite(_sclk, LOW);
    digitalWrite(_mosi, LOW);
    

#if defined (USE_FAST_PINIO)
    clkport     = portOutputRegister(digitalPinToPort(_sclk));
    clkpinmask  = digitalPinToBitMask(_sclk);
    mosiport    = portOutputRegister(digitalPinToPort(_mosi));
    mosipinmask = digitalPinToBitMask(_mosi);
    *clkport   &= ~clkpinmask;
    *mosiport  &= ~mosipinmask;
#endif
  }
  
  rotation  = 0;
  cursor_y  = cursor_x = 0;
  textsize  = 1;
  textcolor = 0xFFFF;

}

void ILI9341_SPI::spiwrite(uint8_t c) {

  //Serial.print("0x"); Serial.print(c, HEX); Serial.print(", ");

  if (hwSPI) {
#if defined (__AVR__)
  #ifndef SPI_HAS_TRANSACTION
    uint8_t backupSPCR = SPCR;
    SPCR = mySPCR;
  #endif
    SPDR = c;
    while(!(SPSR & _BV(SPIF)));
  #ifndef SPI_HAS_TRANSACTION
    SPCR = backupSPCR;
  #endif
#else
    SPI.transfer(c);
#endif
  } else {
#if defined(ESP8266) || defined (ARDUINO_ARCH_ARC32)
    for(uint8_t bit = 0x80; bit; bit >>= 1) {
      if(c & bit) {
	digitalWrite(_mosi, HIGH); 
      } else {
	digitalWrite(_mosi, LOW); 
      }
      digitalWrite(_sclk, HIGH);
      digitalWrite(_sclk, LOW);
    }
#else
    // Fast SPI bitbang swiped from LPD8806 library
    for(uint8_t bit = 0x80; bit; bit >>= 1) {
      if(c & bit) {
	//digitalWrite(_mosi, HIGH); 
	*mosiport |=  mosipinmask;
      } else {
	//digitalWrite(_mosi, LOW); 
	*mosiport &= ~mosipinmask;
      }
      //digitalWrite(_sclk, HIGH);
      *clkport |=  clkpinmask;
      //digitalWrite(_sclk, LOW);
      *clkport &= ~clkpinmask;
    }
#endif
  }
}


void ILI9341_SPI::writecommand(uint8_t c) {
#if defined (USE_FAST_PINIO)
  *dcport &= ~dcpinmask;
  *csport &= ~cspinmask;
#else
  digitalWrite(_dc, LOW);
  digitalWrite(_sclk, LOW);
  digitalWrite(_cs, LOW);
#endif

  spiwrite(c);

#if defined (USE_FAST_PINIO)
  *csport |= cspinmask;
#else
  digitalWrite(_cs, HIGH);
#endif
}


void ILI9341_SPI::writedata(uint8_t c) {
#if defined (USE_FAST_PINIO)
  *dcport |=  dcpinmask;
  *csport &= ~cspinmask;
#else
  digitalWrite(_dc, HIGH);
  digitalWrite(_cs, LOW);
#endif

  spiwrite(c);

#if defined (USE_FAST_PINIO)
  *csport |= cspinmask;
#else
  digitalWrite(_cs, HIGH);
#endif
} 

// Initialization command tables for different LCD controllers
#define TFTLCD_DELAY 0xFF

static const uint8_t ILI9341_SPI_regValues[] PROGMEM = {

      /*
0xCB, 5, 0x39, 0x2C, 0x00, 0x34, 0x02,
0xCF, 3, 0x00, 0XC1, 0X30,
0xE8, 3, 0x85, 0x00, 0x78,
0xEA, 2, 0x00, 0x00,
0xED, 4, 0x64, 0x03, 0X12, 0X81,
0xF7, 1, 0x20,
0xC0, 1, 0x23,
0xC1, 1, 0x10,
0xC5, 2, 0x3e, 0x28,
0xC7, 1, 0x86,
0x36, 1, 0x48,
0x3A, 1, 0x55,
0xB1, 2, 0x00, 0x18,
0xB6, 3, 0x08, 0x82, 0x27,
0x11, 0,
TFTLCD_DELAY, 150, 
0x29, 0,
TFTLCD_DELAY, 150,
0x2c, 0
   */
 


  0xEF,3,0x03,0x80,0x02,
  0xCF,3,0x00,0XC1,0X30, 
  0xED,4,0x64,0x03,0X12,0X81,  
  0xE8,3,0x85,0x00,0x78,
  0xCB,5,0x39,0x2C,0x00,0x34,0x02,  
  0xF7,1,0x20, 
  0xEA,2,0x00,0x00,  
  0xC0,1,0x23,
  0xC1,1,0x10, 
  0xC5,2,0x3e,0x28, 
  0xC7,1,0x86, 
  0x36,1,0x48,
  0x3A,1,0x55,  
  0xB1,2,0x00,0x18, 
  0xB6,3,0x08,0x82,0x27, 
  0xF2,1,0x00, 
  0x26,1,0x01, 
  0xE0,15,0x0F,0x31,0x2B,0x0C,0x0E,0x08,0x4E,0xF1,0x37,0x07,0x10,0x03,0x0E,0x09,0x00,  
  0xE1,15,0x00,0x0E,0x14,0x03,0x11,0x07,0x31,0xC1,0x48,0x08,0x0F,0x0C,0x31,0x36,0x0F,
  0x11, 0,
  TFTLCD_DELAY, 150, 
  0x29,0,
  TFTLCD_DELAY, 150, 
  0x2c, 0

};


void ILI9341_SPI::begin(void) {
  
  reset ();
  
  delay(200);
  
  if (hwSPI) spi_begin();

  uint16_t i = 0;
	while(i < sizeof(ILI9341_SPI_regValues)) {
		uint8_t r = pgm_read_byte(&ILI9341_SPI_regValues[i++]);
		uint8_t len = pgm_read_byte(&ILI9341_SPI_regValues[i++]);
		if(r == TFTLCD_DELAY) {
			delay(len);
		} 
    else {
			//Serial.print("Register $"); Serial.print(r, HEX);  Serial.print(" -> "); 
			writecommand(r);
			for (uint8_t d=0; d<len; d++) {
				uint8_t x = pgm_read_byte(&ILI9341_SPI_regValues[i++]);
				writedata(x);
			  //Serial.print(x, HEX); Serial.print(" ");
			}
      //Serial.println();
		}
	}
  if (hwSPI) spi_end();

}

void ILI9341_SPI::reset(void) {
  if (_rst > 0) {
    digitalWrite(_rst, HIGH);
    delay(5);
    digitalWrite(_rst, LOW);
    digitalWrite(_dc, LOW);
    digitalWrite(_cs, HIGH);
    if(hwSPI) { // Using hardware SPI
      spi_end();
    }
    else {
      digitalWrite(_sclk, LOW);
      digitalWrite(_mosi, LOW);
    }
    delay(20);
    digitalWrite(_rst, HIGH);
    delay(150);
  }
}


void ILI9341_SPI::setAddrWindow(uint16_t x0, uint16_t y0, uint16_t x1, uint16_t y1) {

  writecommand(TFTREGISTER_COLADDRSET); // Column addr set
  writedata(x0 >> 8);
  writedata(x0 & 0xFF);     // XSTART 
  writedata(x1 >> 8);
  writedata(x1 & 0xFF);     // XEND

  writecommand(TFTREGISTER_PAGEADDRSET); // Row addr set
  writedata(y0>>8);
  writedata(y0);     // YSTART
  writedata(y1>>8);
  writedata(y1);     // YEND
  
  writecommand(TFTREGISTER_MEMORYWRITE); // write to RAM

}


void ILI9341_SPI::pushColor(uint16_t color) {
  if (hwSPI) spi_begin();

#if defined(USE_FAST_PINIO)
  *dcport |=  dcpinmask;
  *csport &= ~cspinmask;
#else
  digitalWrite(_dc, HIGH);
  digitalWrite(_cs, LOW);
#endif

  spiwrite(color >> 8);
  spiwrite(color);

#if defined(USE_FAST_PINIO)
  *csport |= cspinmask;
#else
  digitalWrite(_cs, HIGH);
#endif

  if (hwSPI) spi_end();
}

void ILI9341_SPI::drawPixel(int16_t x, int16_t y, uint16_t color) {

  if((x < 0) ||(x >= _width) || (y < 0) || (y >= _height)) return;

  if (hwSPI) spi_begin();
  setAddrWindow(x,y,x,y);

#if defined(USE_FAST_PINIO)
  *dcport |=  dcpinmask;
  *csport &= ~cspinmask;
#else
  digitalWrite(_dc, HIGH);
  digitalWrite(_cs, LOW);
#endif

  spiwrite(color >> 8);
  spiwrite(color);

#if defined(USE_FAST_PINIO)
  *csport |= cspinmask;
#else
  digitalWrite(_cs, HIGH);
#endif

  if (hwSPI) spi_end();
}


void ILI9341_SPI::drawFastVLine(int16_t x, int16_t y, int16_t h,
 uint16_t color) {

  // Rudimentary clipping
  if((x >= _width) || (y >= _height)) return;

  if((y+h-1) >= _height) 
    h = _height-y;

  if (hwSPI) spi_begin();
  setAddrWindow(x, y, x, y+h-1);

  uint8_t hi = color >> 8, lo = color;

#if defined(USE_FAST_PINIO)
  *dcport |=  dcpinmask;
  *csport &= ~cspinmask;
#else
  digitalWrite(_dc, HIGH);
  digitalWrite(_cs, LOW);
#endif

  while (h--) {
    spiwrite(hi);
    spiwrite(lo);
  }

#if defined(USE_FAST_PINIO)
  *csport |= cspinmask;
#else
  digitalWrite(_cs, HIGH);
#endif

  if (hwSPI) spi_end();
}


void ILI9341_SPI::drawFastHLine(int16_t x, int16_t y, int16_t w,
  uint16_t color) {

  // Rudimentary clipping
  if((x >= _width) || (y >= _height)) return;
  if((x+w-1) >= _width)  
    w = _width-x;
    
  if (hwSPI) spi_begin();
  setAddrWindow(x, y, x+w-1, y);

  uint8_t hi = color >> 8, lo = color;
#if defined(USE_FAST_PINIO)
  *dcport |=  dcpinmask;
  *csport &= ~cspinmask;
#else
  digitalWrite(_dc, HIGH);
  digitalWrite(_cs, LOW);
#endif
  while (w--) {
    spiwrite(hi);
    spiwrite(lo);
  }
#if defined(USE_FAST_PINIO)
  *csport |= cspinmask;
#else
  digitalWrite(_cs, HIGH);
#endif
  if (hwSPI) spi_end();
}

void ILI9341_SPI::fillScreen(uint16_t color) {
  fillRect(0, 0,  _width, _height, color);
}

// fill a rectangle
void ILI9341_SPI::fillRect(int16_t x, int16_t y, int16_t w, int16_t h,
  uint16_t color) {

  // rudimentary clipping (drawChar w/big text requires this)
  if((x >= _width) || (y >= _height)) return;
  if((x + w - 1) >= _width)  w = _width  - x;
  if((y + h - 1) >= _height) h = _height - y;

  if (hwSPI) spi_begin();
  setAddrWindow(x, y, x+w-1, y+h-1);

  uint8_t hi = color >> 8, lo = color;

#if defined(USE_FAST_PINIO)
  *dcport |=  dcpinmask;
  *csport &= ~cspinmask;
#else
  digitalWrite(_dc, HIGH);
  digitalWrite(_cs, LOW);
#endif

  for(y=h; y>0; y--) {
    for(x=w; x>0; x--) {
      spiwrite(hi);
      spiwrite(lo);
    }
  }
#if defined(USE_FAST_PINIO)
  *csport |= cspinmask;
#else
  digitalWrite(_cs, HIGH);
#endif

  if (hwSPI) spi_end();
}


// Pass 8-bit (each) R,G,B, get back 16-bit packed color
uint16_t ILI9341_SPI::color565(uint8_t r, uint8_t g, uint8_t b) {
  return ((r & 0xF8) << 8) | ((g & 0xFC) << 3) | (b >> 3);
}


void ILI9341_SPI::setRotation(uint8_t m) {

  rotation = m % 4; // can't be higher than 3
  Adafruit_GFX::setRotation(m);
  if (hwSPI) spi_begin();
  writecommand(TFTREGISTER_MEMCONTROL);
  switch (rotation) {
   case 0:
     writedata(TFTREGISTER_MEMCONTROL_MX | TFTREGISTER_MEMCONTROL_BGR);
     break;
   case 1:
     writedata(TFTREGISTER_MEMCONTROL_MV | TFTREGISTER_MEMCONTROL_BGR);
     break;
  case 2:
    writedata(TFTREGISTER_MEMCONTROL_MY | TFTREGISTER_MEMCONTROL_BGR);
    break;
  case 3:
    writedata(TFTREGISTER_MEMCONTROL_MX | TFTREGISTER_MEMCONTROL_MY | TFTREGISTER_MEMCONTROL_MV | TFTREGISTER_MEMCONTROL_BGR);
    break;
  }
  if (hwSPI) spi_end();
}

/*
void ILI9341_SPI::invertDisplay(boolean i) {
  if (hwSPI) spi_begin();
  writecommand(i ? ILI9341_INVON : ILI9341_INVOFF);
  if (hwSPI) spi_end();
}
*/


// Ditto with the read/write port directions, as above.
uint16_t ILI9341_SPI::readID(void) {

  uint16_t id = readReg(0x00);
  return id;
    
}

uint32_t ILI9341_SPI::readReg(uint8_t r) {
  uint32_t id;
  uint8_t x;

  
  writecommand(r);
  delayMicroseconds(50);

  x = readdata();
  id = x;          // Do not merge or otherwise simplify
  id <<= 8;              // these lines.  It's an unfortunate
  x = readdata();
  id  |= x;        // shenanigans that are going on.
  id <<= 8;              // these lines.  It's an unfortunate
  x = readdata();
  id  |= x;        // shenanigans that are going on.
  id <<= 8;              // these lines.  It's an unfortunate
  x = readdata();
  id  |= x;        // shenanigans that are going on.
  if (r == 0xEF){
    id <<= 8;         // these lines.  It's an unfortunate
	  x = readdata();
	  id  |= x;	  
  }
  delay(150); //stabilization time
  return id;
}


////////// stuff not actively being used, but kept for posterity


uint8_t ILI9341_SPI::spiread(void) {
  uint8_t r = 0;

  if (hwSPI) {
#if defined (__AVR__)
  #ifndef SPI_HAS_TRANSACTION
    uint8_t backupSPCR = SPCR;
    SPCR = mySPCR;
  #endif
    SPDR = 0x00;
    while(!(SPSR & _BV(SPIF)));
    r = SPDR;

  #ifndef SPI_HAS_TRANSACTION
    SPCR = backupSPCR;
  #endif
#else
    r = SPI.transfer(0x00);
#endif

  } else {

    for (uint8_t i=0; i<8; i++) {
      digitalWrite(_sclk, LOW);
      digitalWrite(_sclk, HIGH);
      r <<= 1;
      if (digitalRead(_miso))
	r |= 0x1;
    }
  }
  //Serial.print("read: 0x"); Serial.print(r, HEX);
  
  return r;
}

 uint8_t ILI9341_SPI::readdata(void) {
   digitalWrite(_dc, HIGH);
   digitalWrite(_cs, LOW);
   uint8_t r = spiread();
   digitalWrite(_cs, HIGH);
   
   return r;
}
 

uint8_t ILI9341_SPI::readcommand8(uint8_t c, uint8_t index) {
   if (hwSPI) spi_begin();
   digitalWrite(_dc, LOW); // command
   digitalWrite(_cs, LOW);
   spiwrite(0xD9);  // woo sekret command?
   digitalWrite(_dc, HIGH); // data
   spiwrite(0x10 + index);
   digitalWrite(_cs, HIGH);

   digitalWrite(_dc, LOW);
   digitalWrite(_sclk, LOW);
   digitalWrite(_cs, LOW);
   spiwrite(c);
 
   digitalWrite(_dc, HIGH);
   uint8_t r = spiread();
   digitalWrite(_cs, HIGH);
   if (hwSPI) spi_end();
   return r;
}

